"""
These functions resolve dispatcher element instances from supplied dictionaries.
"""
from pydantic import ValidationError
import whendo.core.action as action
import whendo.core.scheduler as scheduler
import whendo.core.util as util

def resolve_action(dictionary:dict, check_for_found_class:bool=True):
    return util.resolve_instance(action.Action, dictionary, check_for_found_class=check_for_found_class)

def resolve_scheduler(dictionary:dict, check_for_found_class:bool=True):
    return util.resolve_instance(scheduler.Scheduler, dictionary, check_for_found_class=check_for_found_class)

def resolve_file_pathe(dictionary:dict, check_for_found_class:bool=True):
    try:
        return util.FilePathe(**dictionary)
    except ValidationError as error:
        if check_for_found_class:
            raise error
        else:
            return dictionary



"""
These classes perform simple operations on pins.
"""
try:
    import RPi.GPIO as GPIO
except:
    import Mock.GPIO as GPIO

from whendo.core.action import Action

class SetPin(Action):
    """
    For the specified pin, sets to HIGH if on=True, to LOW if on=False.
    """
    pin:int
    on:bool

    def execute(self, tag:str=None, scheduler_info:dict=None):
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup(self.pin, GPIO.OUT)
        GPIO.output(self.pin, GPIO.HIGH if self.on else GPIO.LOW)

class TogglePin(Action):
    """
    For the specified pin, sets to HIGH if LOW, to LOW if HIGH.
    """
    pin:int

    def execute(self, tag:str=None, scheduler_info:dict=None):
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup(self.pin, GPIO.OUT)
        GPIO.output(self.pin, not GPIO.input(self.pin))

class Cleanup(Action):
    """
    Clean up the pins. See the docs for GPIO.cleanup().
    """
    cleanup:str='cleanup'

    def execute(self, tag:str=None, scheduler_info:dict=None):
        GPIO.cleanup()


"""
Instances of this class manage dispatchers in putative dispatcher-ready, network-visible hosts.

A host will at least have ip address and port as properties
"""
from pydantic import BaseModel
from typing import List

class Fleet(BaseModel):
    host:FleetHost
    fleet_hosts:List[FleetHost]=[]

    def collect_hosts(self):
        """
        populate hosts with locatable fleet hosts
        """
        pass

class FleetHost(BaseModel):
    name:str
    description:str
    host:str
    host_ports:List[HostPort]=[]

class HostPort(BaseModel):
    name:str
    port:int


"""
This module is the top of the path tree. Establishes all sub-path routers.

It is used for command line invocation.
"""
from fastapi import FastAPI
from whendo.core.dispatcher import DispatcherSingleton
from whendo.core.continuous import Continuous
from whendo.api.router import actions, schedulers, dispatcher, jobs
from whendo.api.shared import set_dispatcher
import whendo.log.init

app = FastAPI()

@app.get('/')
async def root():
    return 'whengo API server started'

dispatcher_instance = DispatcherSingleton.get()

app.include_router(set_dispatcher(actions.router, dispatcher_instance))
app.include_router(set_dispatcher(schedulers.router, dispatcher_instance))
app.include_router(set_dispatcher(dispatcher.router, dispatcher_instance))
app.include_router(set_dispatcher(jobs.router, dispatcher_instance))


import logging
import os
import whendo.core.util as util

logging_config = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default_handler': {
            'class': 'logging.FileHandler',
            'level': 'INFO',
            'formatter': 'standard',
            'filename': os.path.join(util.Dirs.log_dir(), 'job.log'),
            'encoding': 'utf8'
        },
    },
    'loggers': {
        '': {
            'handlers': ['default_handler'],
            'level': 'DEBUG',
            'propagate': False
        }
    }
}
logging.config.dictConfig(logging_config)


from fastapi import APIRouter, status, Depends
from whendo.api.shared import return_success, raised_exception, get_dispatcher
from whendo.core.resolver import resolve_scheduler

router = APIRouter(
    prefix='',
    tags=['Schedulers']
)

@router.get('/schedulers/{scheduler_name}/actions/{action_name}', status_code=status.HTTP_200_OK)
def schedule_action(scheduler_name:str, action_name:str):
    try:
        get_dispatcher(router).schedule_action(scheduler_name=scheduler_name, action_name=action_name)
        return return_success(f"action ({action_name}) was successfully scheduled ({scheduler_name})")
    except Exception as e:
        raise raised_exception(f"failed to schedule ({scheduler_name}) action ({action_name})", e)

@router.get('/schedulers/{scheduler_name}', status_code=status.HTTP_200_OK)
def get_scheduler(scheduler_name:str):
    try:
        scheduler = get_dispatcher(router).get_scheduler(scheduler_name=scheduler_name)
        return return_success(scheduler)
    except Exception as e:
        raise raised_exception(f"failed to retrieve the scheduler ({scheduler_name})", e)

@router.post('/schedulers/{scheduler_name}', status_code=status.HTTP_200_OK)
def add_scheduler(scheduler_name:str, scheduler=Depends(resolve_scheduler)):
    try:
        assert scheduler, f"couldn't resolve class for scheduler ({scheduler_name})"
        get_dispatcher(router).add_scheduler(scheduler_name=scheduler_name, scheduler=scheduler)
        return return_success(f"scheduler ({scheduler_name}) was successfully added")
    except Exception as e:
        raise raised_exception(f"failed to add scheduler ({scheduler_name})", e)

@router.put('/schedulers/{scheduler_name}', status_code=status.HTTP_200_OK)
def set_scheduler(scheduler_name:str, scheduler=Depends(resolve_scheduler)):
    try:
        assert scheduler, f"couldn't resolve class for scheduler ({scheduler_name})"
        get_dispatcher(router).set_scheduler(scheduler_name=scheduler_name, scheduler=scheduler)
        return return_success(f"scheduler ({scheduler_name}) was successfully updated")
    except Exception as e:
        raise raised_exception(f"failed to update scheduler ({scheduler_name})", e)

@router.delete('/schedulers/{scheduler_name}', status_code=status.HTTP_200_OK)
def delete_scheduler(scheduler_name:str):
    try:
        get_dispatcher(router).delete_scheduler(scheduler_name=scheduler_name)
        return return_success(f"scheduler ({scheduler_name}) was successfully deleted")
    except Exception as e:
        raise raised_exception(f"failed to delete scheduler ({scheduler_name})", e)

@router.get('/schedulers/{scheduler_name}/execute', status_code=status.HTTP_200_OK)
def execute_scheduler_actions(scheduler_name:str):
    try:
        result = get_dispatcher(router).execute_scheduler_actions(scheduler_name=scheduler_name)
        return return_success({'msg': f"scheduler ({scheduler_name}) actions were successfully executed", 'result':result})
    except Exception as e:
        raise raised_exception(f"failed to execute scheduler ({scheduler_name}) actions", e)

@router.get('/schedulers/{scheduler_name}/unschedule', status_code=status.HTTP_200_OK)
def unschedule_scheduler(scheduler_name:str):
    try:
        get_dispatcher(router).unschedule_scheduler(scheduler_name=scheduler_name)
        return return_success(f"scheduler ({scheduler_name}) was successfully unscheduled")
    except Exception as e:
        raise raised_exception(f"failed to unschedule scheduler ({scheduler_name})", e)

@router.get('/schedulers/{scheduler_name}/reschedule', status_code=status.HTTP_200_OK)
def reschedule_scheduler(scheduler_name:str):
    try:
        get_dispatcher(router).reschedule_scheduler(scheduler_name=scheduler_name)
        return return_success(f"scheduler ({scheduler_name}) was successfully rescheduled")
    except Exception as e:
        raise raised_exception(f"failed to reschedule scheduler ({scheduler_name})", e)

@router.get('/schedulers/reschedule_all', status_code=status.HTTP_200_OK)
def reschedule_all_schedulers():
    try:
        get_dispatcher(router).reschedule_all_schedulers()
        return return_success("all schedulers were successfully unscheduled")
    except Exception as e:
        raise raised_exception("failed to unschedule all schedulers", e)


import pytest
import time
from whendo.core.dispatcher import Dispatcher
from test.fixtures import friends

def test_schedule_action(friends):
    """
    Tests Dispatcher and Continuous objects running a scheduled action.
    """
    dispatcher, scheduler, action = friends()

    dispatcher.add_action('foo', action)
    dispatcher.add_scheduler('bar', scheduler)
    dispatcher.schedule_action('bar', 'foo')

    continuous = dispatcher.get_continuous()
    continuous.run_continuously()
    time.sleep(4)
    continuous.stop_running_continuously()
    continuous.clear()

    line = None
    with open(action.file, 'r') as fid:
        line = fid.readline()
    assert line is not None and type(line) is str and len(line) > 0

def test_unschedule_action(friends):
    """
    Tests unscheduling an action.
    """
    dispatcher, scheduler, action = friends()

    dispatcher.add_action('foo', action)
    dispatcher.add_scheduler('bar', scheduler)
    dispatcher.schedule_action('bar', 'foo')
    dispatcher.unschedule_action('foo')

    continuous = dispatcher.get_continuous()
    assert continuous.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == continuous.job_count()

def test_reschedule_action(friends):
    """
    Tests unscheduling and then rescheduling an action.
    """
    dispatcher, scheduler, action = friends()
    continuous = dispatcher.get_continuous()
    assert continuous.job_count() == 0

    dispatcher.add_action('foo', action)
    dispatcher.add_scheduler('bar', scheduler)
    dispatcher.schedule_action('bar', 'foo')
    dispatcher.reschedule_action('foo')

    continuous.run_continuously()
    time.sleep(4)
    continuous.stop_running_continuously()
    continuous.clear()

    line = None
    with open(action.file, 'r') as fid:
        line = fid.readline()

    assert line is not None and type(line) is str and len(line) > 0

def test_reschedule_action_2(friends, tmp_path):
    """
    Tests unscheduling and then rescheduling an action.
    """
    dispatcher, scheduler, action = friends()
    continuous = dispatcher.get_continuous()
    assert continuous.job_count() == 0

    dispatcher.add_action('foo', action)
    dispatcher.add_scheduler('bar', scheduler)
    dispatcher.schedule_action('bar', 'foo')
    stored_action = dispatcher.get_action('foo')
    stored_action.file = str(tmp_path / 'output2.txt')
    dispatcher.set_action('foo', stored_action)
    dispatcher.reschedule_action('foo')

    continuous.run_continuously()
    time.sleep(4)
    continuous.stop_running_continuously()
    continuous.clear()

    line = None
    with open(stored_action.file, 'r') as fid:
        line = fid.readline()
    assert line is not None and type(line) is str and len(line) > 0

def test_unschedule_scheduler(friends):
    """
    Tests unscheduling a scheduler.
    """
    dispatcher, scheduler, action = friends()
    continuous = dispatcher.get_continuous()
    assert continuous.job_count() == 0

    dispatcher.add_action('foo', action)
    dispatcher.add_scheduler('bar', scheduler)
    dispatcher.schedule_action('bar', 'foo')
    
    assert continuous.job_count() == 1

    dispatcher.unschedule_scheduler('bar')

    assert continuous.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == continuous.job_count()

    # make sure that bar and foo remain
    assert dispatcher.get_scheduler('bar')
    assert dispatcher.get_action('foo')


def test_clear_dispatcher(friends):
    """
    Tests clearing a dispatcher.
    """
    dispatcher, scheduler, action = friends()
    continuous = dispatcher.get_continuous()
    assert continuous.job_count() == 0

    dispatcher.add_action('foo', action)
    dispatcher.add_scheduler('bar', scheduler)
    dispatcher.schedule_action('bar', 'foo')
    assert continuous.job_count() == 1
    
    dispatcher.clear_all()
    assert continuous.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == continuous.job_count()

    # make sure that bar and foo are Gone
    assert dispatcher.get_scheduler('bar') is None
    assert dispatcher.get_action('foo') is None

def test_saved_dir_1(tmp_path):
    saved_dir = str(tmp_path)
    dispatcher = Dispatcher()
    dispatcher.set_saved_dir(saved_dir=saved_dir)
    assert dispatcher.get_saved_dir() == saved_dir

def test_saved_dir_2(tmp_path):
    saved_dir = str(tmp_path)
    dispatcher = Dispatcher(saved_dir=saved_dir)
    assert dispatcher.get_saved_dir() == saved_dir


from typing import Dict, Any, Optional
from whendo.core.util import PP
from whendo.core.action import Action, ActionPayload

class FileHeartbeat(Action):
    """
    This class appends status information to a file. This information can include
    a dictionary [xtra] supplied at instantiation.
    """
    file: str
    xtra: Optional[Dict[str, Any]]=None

    def execute(self, tag:str=None, scheduler_info:Dict[str, Any]=None):
        payload = ActionPayload.build(action_info=self.info(), scheduler_info=scheduler_info)
        payload['tag'] = tag if tag else 'N/A'
        if self.xtra:
            payload['xtra'] = self.xtra # dictionaries are sorted by key. Nice to have extra information at the bottom.
        with open(self.file, "a") as outfile:
            PP.pprint(payload, stream=outfile)
            outfile.write('\n')
        return {'outcome':'file appended', 'action':self.info()}
    def set_xtra(self, xtra:Dict[str, Any]=None):
        self.xtra = xtra
        return self
    def get_xtra(self):
        return self.xtra


"""
pytest fixtures for unit testing
"""

import random
import pytest
from whendo.core.dispatcher import Dispatcher
from whendo.core.actions.file_action import FileHeartbeat
from whendo.core.scheduler import TimelyScheduler
from whendo.core.continuous import Continuous
from whendo.api import main_temp
from test.uvicorn_server import UvicornTestServer

@pytest.fixture
def friends(tmp_path):
    """ returns a tuple of useful test objects """
    def stuff():
        # want a fresh tuple from the fixture
        saved_dir = str(tmp_path)
        output_file = str(tmp_path / 'output.txt')
        dispatcher = Dispatcher(saved_dir=saved_dir)
        dispatcher.set_continuous(Continuous())
        action = FileHeartbeat(file=output_file)
        scheduler = TimelyScheduler(interval=1)

        return dispatcher, scheduler, action
    return stuff

@pytest.fixture
def port():
    """ returns a random port number in case the unit test system runs test in parallel """
    return random.randint(5001,7999)

@pytest.fixture
def host():
    """ the host used in unit tests """
    return '127.0.0.1'

@pytest.fixture
async def startup_and_shutdown_uvicorn(host, port):
    """Start server as test fixture and tear down after test"""
    uvicorn_server = UvicornTestServer(app=main_temp.app, host=host, port=port)
    await uvicorn_server.up()
    yield
    await uvicorn_server.down()

@pytest.fixture
def base_url(host, port):
    """ base url for unit test invocation of api """
    return f"http://{host}:{port}"


from pydantic import BaseModel
import requests
from whendo.core.action import Action
from whendo.core.scheduler import Scheduler
from whendo.core.resolver import resolve_action, resolve_scheduler, resolve_file_pathe
from whendo.core.util import FilePathe
from whendo.core.dispatcher import Dispatcher

class Client(BaseModel):
    host:str='127.0.0.1'
    port:int=8000

    # /dispatcher
    def load_dispatcher(self):
        return Dispatcher.resolve(self.get(f"/dispatcher/load"))
    def save_dispatcher(self):
        return self.get("/dispatcher/save")
    def clear_dispatcher(self):
        return self.get("/dispatcher/clear")
    def load_dispatcher_from_name(self, name:str):
        return Dispatcher.resolve(self.get(f"/dispatcher/load_from_name/{name}"))
    def save_data_to_name(self, name:str):
        return self.get(f"/dispatcher/save_to_name/{name}")
    def get_saved_dir(self):
        return resolve_file_pathe(self.get("/dispatcher/saved_dir"))
    def set_saved_dir(self, saved_dir:FilePathe):
        return self.put("/dispatcher/saved_dir", saved_dir)

    # /actions
    def get_action(self, action_name:str):
        return resolve_action(self.get(f"/actions/{action_name}"))
    def add_action(self, action_name:str, action:Action):
        return self.post(f"/actions/{action_name}", action)
    def set_action(self, action_name:str, action:Action):
        return self.put(f"/actions/{action_name}", action)
    def delete_action(self, action_name:str):
        return self.delete(f"/actions/{action_name}")
    def execute_action(self, action_name:str):
        return self.get(f"/actions/{action_name}/execute")
    def unschedule_action(self, action_name:str):
        return self.get(f"/actions/{action_name}/unschedule")
    def reschedule_action(self, action_name:str):
        return self.get(f"/actions/{action_name}/reschedule")

    # /schedulers
    def schedule_action(self, scheduler_name:str, action_name:str):
        return self.get(f"/schedulers/{scheduler_name}/actions/{action_name}")
    def get_scheduler(self, scheduler_name:str):
        return resolve_scheduler(self.get(f"/schedulers/{scheduler_name}"))
    def add_scheduler(self, scheduler_name:str, scheduler:Scheduler):
        return self.post(f"/schedulers/{scheduler_name}", scheduler)
    def set_scheduler(self, scheduler_name:str, scheduler:Scheduler):
        return self.put(f"/schedulers/{scheduler_name}", scheduler)
    def delete_scheduler(self, scheduler_name:str):
        return self.delete(f"/schedulers/{scheduler_name}")
    def unschedule_scheduler(self, scheduler_name:str):
        return self.get(f"/schedulers/{scheduler_name}/unschedule")
    def reschedule_all_schedulers(self):
        return self.get(f"/schedulers/reschedule_all")
    def execute_scheduler_actions(self, scheduler_name:str):
        return self.get(f"/schedulers/{scheduler_name}/execute")

    # /jobs
    def run_jobs(self):
        return self.get(f"/jobs/run")
    def stop_jobs(self):
        return self.get(f"/jobs/stop")
    def jobs_are_running(self):
        return self.get(f"/jobs/are_running")
    def job_count(self):
        return self.get(f"/jobs/count")
    def clear_jobs(self):
        return self.get(f"/jobs/clear")

    # verbs
    def get(self, path:str, data=None):
        response = requests.get(self.cmd(path), data)
        assert response.status_code == 200, response.text
        return response.json()
    def put(self, path:str, data:BaseModel):
        response = requests.put(self.cmd(path), data.json())
        assert response.status_code == 200, response.text
        return response.json()
    def post(self, path:str, data:BaseModel):
        response = requests.post(self.cmd(path), data.json())
        assert response.status_code == 200, response.text
        return response.json()
    def delete(self, path:str):
        response = requests.delete(self.cmd(path))
        assert response.status_code == 200, response.text
        return response.json()
    def cmd(self, path:str):
        return f"http://{self.host}:{self.port}{path}"



from fastapi import APIRouter, status
from whendo.api.shared import return_success, raised_exception, get_dispatcher

router = APIRouter(
    prefix="/jobs",
    tags=['Jobs']
)

@router.get('/run', status_code=status.HTTP_200_OK)
def start_scheduled_jobs():
    try:
        assert not get_dispatcher(router).jobs_are_running(), "should not be running when trying to start"
        get_dispatcher(router).run_jobs()
        return return_success("started running")
    except Exception as e:
        raise raised_exception("failed to start running", e)

@router.get('/stop', status_code=status.HTTP_200_OK)
def stop_scheduled_jobs():
    try:
        assert get_dispatcher(router).jobs_are_running(), "should be running when trying to stop"
        get_dispatcher(router).stop_jobs()
        return return_success("stopped running")
    except Exception as e:
        raise raised_exception("failed to stop running", e)

@router.get('/count', status_code=status.HTTP_200_OK)
def job_count():
    try:
        return return_success({'job_count':get_dispatcher(router).job_count()})
    except Exception as e:
        raise raised_exception("failed to get job count", e)

@router.get('/are_running', status_code=status.HTTP_200_OK)
def jobs_are_running():
    try:
        return return_success({'job_count':get_dispatcher(router).jobs_are_running()})
    except Exception as e:
        raise raised_exception("failed to get job count", e)

@router.get('/clear', status_code=status.HTTP_200_OK)
def clear():
    try:
        return return_success({'job_count':get_dispatcher(router).clear_jobs()})
    except Exception as e:
        raise raised_exception("failed to get job count", e)


from enum import Enum
from pprint import PrettyPrinter
from sys import stdout
# import netifaces
from datetime import datetime
from typing import Callable
import os
from pydantic import BaseModel
from typing import List, Dict, Any
from pathlib import Path
from threading import RLock

# Enums

class HttpVerb(str, Enum):
    """
    usage:
        HttpVerb.get
    """
    get = 'get'
    put = 'put'
    patch = 'patch'
    post = 'post'
    delete = 'delete'

class TimeUnit(str, Enum):
    """
    usage:
        TimeUnit.minute
    """
    day = 'day'
    hour = 'hour'
    minute = 'minute'
    second = 'second'

# functions

def all_visible_subclasses(klass):
    def helper(klas, result):
        subklases = klas.__subclasses__()
        if len(subklases) > 0:
            result.update(subklases)
            for subklas in subklases:
                helper(subklas, result)
    result = set()
    helper(klass, result)
    return result

def key_strings_from_dict(dictionary:Dict[str, Any]):
    return set(x for x in dictionary)

def key_strings_from_class(klass):
    return set(x for x in klass.__fields__.keys())

def find_class(klass, dictionary:Dict[str, Any]):
    dictionary_keys = key_strings_from_dict(dictionary)
    classes = all_visible_subclasses(klass)
    classes.add(klass) # the top class might not be abstract
    min_count = 100
    found_class = None
    for clas in classes:
        class_keys = key_strings_from_class(clas)
        if len(dictionary_keys - class_keys) == 0: # make sure class has enough fields
            count = len(class_keys)
            if count<min_count: # pick class with fewest fields, resulting in tightest conformance with dictionary
                found_class = clas
                min_count = count
    return found_class

def resolve_instance(klass, dictionary:Dict[str, Any], check_for_found_class:bool=True):
    """
    if an (improper) subclass is found that maps to the supplied dictionary, the dictionary
    is converted to an instance of that class. If the dictionary contains lists of dictionaries,
    those dictionaries are recursively resolved.
    """
    found_class = find_class(klass, dictionary)
    if found_class is None:
        if check_for_found_class:
            raise NameError(f"could not resolve dictionary ({dictionary}) to a subclass of ({klass})")
        else:
            return dictionary
    else:
        # resolve singleton dictionary elements; not constrained to producing instance of klass subclass
        for (key, value) in dictionary.items():
            if isinstance(value, dict):
                dictionary[key] = resolve_instance(klass, dictionary=value, check_for_found_class=False)
        # resolve list containing all dictionary elements; likelihood of non-klass in this circumstance
        # is very small, therefore more strict than for singletons
        for (key, value) in dictionary.items():
            if isinstance(value, list):
                if all(isinstance(element, dict) for element in value):
                    dictionary[key] = list(resolve_instance(klass, dictionary=element) for element in value)
        return found_class(**dictionary)

def object_info(obj):
    return {
        'class': f"{obj.__class__.__module__}.{obj.__class__.__name__}",
        'instance': obj
    }

# classes

class Now:
    @classmethod
    def s(cls):
        now = cls.dt()
        return f"{now.strftime('%H:%M:%S')} on {now.strftime('%Y-%m-%d')}"
    @classmethod
    def dt(cls):
        return datetime.now()
    @classmethod
    def t(cls):
        return cls.dt().time()

class Output:
    """
    usage:
    Output.pprint(dictionary={'a', 'abc'}, lev_fil=('debug', Output.debug()))
    """
    @classmethod
    def debug(cls):
        return 'debug', Dirs.output_dir() + 'output.txt'

    @classmethod
    def pprint(cls, lev_fil:tuple[str,str], dictionary:dict):
        level, file = lev_fil
        something = {'level':level}
        something.update(dictionary)
        with open(file, 'a') as outfile:
            PP.pprint(something, stream=outfile)

class PP:
    """
    usage:
        PP.pprint(json_ish_object)
    note:
        could be a function, but PP asserts its global nature
    """
    @classmethod
    def pprint(cls, dictionary, indent=2, stream=stdout):
        PrettyPrinter(indent=indent, stream=stream).pprint(dictionary)

class IP:
    """
    usage:
        IP.addr
    note:
        singleton
    lineage:
        from netifaces import interfaces, ifaddresses, AF_INET
        for ifaceName in interfaces():
            addresses = [i['addr'] for i in ifaddresses(ifaceName).setdefault(AF_INET, [{'addr':'No IP addr'}] )]
            print ('%s: %s' % (ifaceName, ', '.join(addresses)))
    """
    local = '127.0.0.1'
    addr = "need a fix for linux"#netifaces.ifaddresses('en0').setdefault(netifaces.AF_INET)[0]['addr']
    @classmethod
    def reset(cls):
        cls.addr = "need a fix for linux"#netifaces.ifaddresses('en0').setdefault(AF_INET)[0]['addr']


class Dirs:
    """
    This class computes directory paths for saved, output and log files. It creates the directories if absent.
    """
    @classmethod
    def saved_dir(cls, root_path:Path=None):
        return cls.assure_dir(cls.compute_dir(label='saved', root_path=root_path))
    @classmethod
    def output_dir(cls, root_path:Path=None):
        return cls.assure_dir(cls.compute_dir(label='output', root_path=root_path))
    @classmethod
    def log_dir(cls, root_path:Path=None):
        return cls.assure_dir(cls.compute_dir(label='log', root_path=root_path))

    @classmethod
    def compute_dir(cls, label:str, root_path:Path=None):
        path = root_path if root_path else Path.home()
        return str(path / f".whendo/{label}") + '/'
    @classmethod
    def assure_dir(cls, assured_dir:str):
        if not os.path.exists(assured_dir):
            os.makedirs(assured_dir)
        return assured_dir

class FilePathe(BaseModel):
    path: str

class SharedRO():
    """
    This class provides in-memory shared data so that actions can communicate with each other during
    an api server's lifetime. Not meant to be persisted. Persisting shared data is a different feature.

    The point of this class is to treat the initializing dictionary as read-only data. Subsequent
    accesses always produce the same result. The contract has only one access point for the contained
    dictionary, the data_copy method.

    Instances not meant to be created directly. Use Shared_ROs.get(...) instead.

    Usage:
        sh = Shared_ROs.get('foo', some_dictionary)
        assert sh.get('foo') == some_dictionary # not identity comparison
    """
    def __init__(self, dictionary:dict={}):
        self.data = dictionary.copy()

    def data_copy(self):
        return self.data.copy()

class SharedRW(SharedRO):
    """
    This class provides in-memory shared data so that actions can communicate with each other during
    an api server's lifetime. Not meant to be persisted. Persisting shared data is a different feature.
    
    The point of this class is to treat the initializing dictionary as writeable data. Review the
    apply method, which implements a critical section with a threading.RLock. The supplied Callable
    operates on a copy of the contained dictionary.

    Instances not meant to be created directly. Use Shared_RWs.get(...) instead.

    Usage:
        sh = Shared_RWs.get('foo')
        def update(d:dict):
            d['a'] = 1
            return d['a']
        result = sh.apply(update)
        assert 1 == result
        assert sh.data.copy()['a'] == 1
    """
    def __init__(self, dictionary:dict={}):
        super().__init__(dictionary=dictionary)
        self.lock = RLock()
    
    def apply(self, callable:Callable):
        """
        Transactionally applies callable to a copy of the data. A failure within
        the callable won't corrupt the dictionary.
        """
        with self.lock:
            copy = self.data_copy()
            result = callable(copy)
            self.data = copy
            return result

class SharedROs:
    """
    A dictionary of instances of SharedRO's.
    """
    singletons:Dict[str, SharedRO] = {}

    @classmethod
    def get(cls, label:str, dictionary:dict) -> SharedRO:
        if label not in cls.singletons:
            cls.singletons[label] = SharedRO(dictionary=dictionary)
        return cls.singletons[label]
    
    @classmethod
    def key_set(cls):
        return set(cls.singletons.keys())

class SharedRWs:
    """
    A dictionary of instances of SharedWR's.
    """
    singletons:Dict[str, SharedRW] = {}

    @classmethod
    def get(cls, label:str, dictionary:dict={}) -> SharedRW:
        if label not in cls.singletons:
            cls.singletons[label] = SharedRW(dictionary=dictionary)
        return cls.singletons[label]
    
    @classmethod
    def key_set(cls):
        return set(cls.singletons.keys())


"""
This script runs the api server with optionally supplied host and port.
"""
import argparse
import uvicorn

"""
import the main script that creates the FastAPI instance (main.app).
"""
from whendo.api import main

"""
parse command line arguments
"""
parser = argparse.ArgumentParser()
parser.add_argument('--host', type=str, default='127.0.0.1', dest='host')
parser.add_argument('--port', type=int, default=8000, dest='port')
parser.add_argument('--workers', type=int, default=1, dest='workers')
args = parser.parse_args()

"""
uvicorn is the ASGI server that runs the api specified with FastAPI.
"""
uvicorn.run(main.app, host=args.host, port=args.port, workers=args.workers)



"""
This module is the top of the path tree. Establishes all sub-path routers.

It is used for unit testing.
"""
from fastapi import FastAPI
from whendo.core.dispatcher import Dispatcher
from whendo.core.continuous import Continuous
from whendo.api.router import actions, schedulers, dispatcher, jobs
from whendo.api.shared import set_dispatcher

app = FastAPI()

@app.get('/')
async def root():
    return 'whengo API server started (unit test)'

dispatcher_instance = Dispatcher()
dispatcher_instance.set_continuous(Continuous())

app.include_router(set_dispatcher(actions.router, dispatcher_instance))
app.include_router(set_dispatcher(schedulers.router, dispatcher_instance))
app.include_router(set_dispatcher(dispatcher.router, dispatcher_instance))
app.include_router(set_dispatcher(jobs.router, dispatcher_instance))


from pydantic import BaseModel
from typing import Dict, Any
from whendo.core.util import PP, IP, Now, object_info

class Action(BaseModel):
    """
    Actions get something done.
    """
    def execute(self, tag:str=None, scheduler_info:dict=None):
        """
        This method attempts to do something useful and return something useful
        """
        return {'result':'something useful happened'}

    def info(self):
        return object_info(self)
    
    def flat(self):
        return self.json()

# auxilliary
   
class ActionPayload:
    """
    This class produces a status-oriented payload, including action and scheduler info.
    """
    @classmethod
    def build(cls, action_info:Dict[str, Any], scheduler_info:Dict[str, Any]=None):
        payload = {}
        payload.update({'action_info':action_info})
        if scheduler_info:
            payload.update({'scheduler_info':scheduler_info})
        payload.update(cls.action_host())
        payload.update(cls.action_time())
        return payload
    @classmethod
    def action_host(cls):
        return {'action_host': IP.addr}
    @classmethod
    def action_time(cls):
        return {'action_time':Now.s()}


"""
This script sets up a library.

Usage:

    library creation...

        python setup.py bdist_wheel

        [stores .whl file in ./dist directory]

            ./dist/whendo-0.0.1-py3-none-any.whl

    library installation with example file...

        pip install /path/to/whendo-0.0.1-py3-none-any.whl

"""
from setuptools import setup, find_packages, find_namespace_packages

# To use a consistent encoding
from codecs import open
import os

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

# This call to setup() does all the work
setup(
    name="whendo",
    version="0.0.1",
    description="action scheduling library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/electronhead/whendo",
    author="Gary Beaver",
    author_email="gary.beaver@electronhead.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        # "Programming Language :: Python",
        # "Programming Language :: Python :: 3",
        # "Programming Language :: Python :: 3.6",
        # "Programming Language :: Python :: 3.7",
        # "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    packages=get_packages('whendo'),
    # packages=find_packages(exclude=['test', 'core', 'api', 'sdk']),
    install_requires=["uvicorn", "fastapi", "pydantic", "schedule", "requests", "netifaces", "Mock.GPIO", "RPi.GPIO"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "httpx", "pytest-asyncio", "asyncio"],
    test_suite="test"
)


"""
These test cases run the API within each test function call, allowing API http calls
to a live and albeit short-lived whendo server.

This class and the fixture, startup_and_shutdown_uvicorn, rely on asynchronous processing.
"""
import time
import pytest
from pydantic import BaseModel
from httpx import AsyncClient
from test.client_async import ClientAsync
from whendo.core.action import Action
from whendo.core.actions.file_action import FileHeartbeat
from whendo.core.actions.logic_action import ListAction, ListOpMode
from whendo.core.scheduler import TimelyScheduler, Scheduler
from whendo.core.util import FilePathe, resolve_instance
from test.fixtures import port, host, startup_and_shutdown_uvicorn
from whendo.core.resolver import resolve_action, resolve_scheduler, resolve_file_pathe

@pytest.mark.asyncio
async def test_client_1(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """
    add action and scheduler
    """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    output_file = str(tmp_path / 'output.txt')
    action = FileHeartbeat(file=output_file)
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo', action=action)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)

@pytest.mark.asyncio
async def test_client_2(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """
    add action and scheduler and then schedule the action
    """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    output_file = str(tmp_path / 'output.txt')
    action = FileHeartbeat(file=output_file)
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo', action=action)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo', scheduler_name='bar')
    await assert_job_count(client=client, n=1)

@pytest.mark.asyncio
async def test_client_3(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """
    add action and scheduler, run continuous, and then make sure the action produced file output
    """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    output_file = str(tmp_path / 'output.txt')
    action = FileHeartbeat(file=output_file)
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo', action=action)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo', scheduler_name='bar')
    await assert_job_count(client=client, n=1)

    await run_and_stop_jobs(client=client, pause=4)

    lines = None
    with open(action.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1

@pytest.mark.asyncio
async def test_client_logic_action(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """ Run two actions within one action. """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    action1 = FileHeartbeat(file=str(tmp_path / 'output1.txt'))
    action2 = FileHeartbeat(file=str(tmp_path / 'output2.txt'))
    action3 = ListAction(op_mode=ListOpMode.ALL, action_list=[action1, action2])
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo', action=action3)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo', scheduler_name='bar')
    await assert_job_count(client=client, n=1)

    await run_and_stop_jobs(client=client, pause=4)

    lines = None
    with open(action1.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1

    lines = None
    with open(action2.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1

@pytest.mark.asyncio
async def test_set_action_1(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """ unschedule an action. """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    action1 = FileHeartbeat(file=str(tmp_path / 'output1.txt'))
    action2 = FileHeartbeat(file=str(tmp_path / 'output2.txt'))
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo', action=action1)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo', scheduler_name='bar')
    await assert_job_count(client=client, n=1)

    await run_and_stop_jobs(client=client, pause=4)

    lines = None
    with open(action1.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1

    await set_action(client=client, action_name='foo', action=action2)
    await reschedule_action(client=client, action_name='foo')
    await assert_job_count(client=client, n=1)

    await run_and_stop_jobs(client=client, pause=4)

    lines = None
    with open(action2.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1

@pytest.mark.asyncio
async def test_unschedule_action_1(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """ unschedule an action. """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    action1 = FileHeartbeat(file=str(tmp_path / 'output1.txt'))
    action2 = FileHeartbeat(file=str(tmp_path / 'output2.txt'))
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo1', action=action1)
    await add_action(client=client, action_name='foo2', action=action2)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo1', scheduler_name='bar')
    await schedule_action(client=client, action_name='foo2', scheduler_name='bar')
    await assert_job_count(client=client, n=2)

    await unschedule_action(client=client, action_name='foo1')
    await assert_job_count(client=client, n=1)
    await get_action(client=client, action_name='foo1')
    await get_action(client=client, action_name='foo2')
    await get_scheduler(client=client, scheduler_name='bar')

@pytest.mark.asyncio
async def test_unschedule_action_2(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """ unschedule an action. make sure both schedulers are affected."""
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    action1 = FileHeartbeat(file=str(tmp_path / 'output1.txt'))
    action2 = FileHeartbeat(file=str(tmp_path / 'output2.txt'))
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo1', action=action1)
    await add_action(client=client, action_name='foo2', action=action2)
    await add_scheduler(client=client, scheduler_name='bar1', scheduler=scheduler)
    await add_scheduler(client=client, scheduler_name='bar2', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo1', scheduler_name='bar1')
    await schedule_action(client=client, action_name='foo1', scheduler_name='bar2')
    await schedule_action(client=client, action_name='foo2', scheduler_name='bar1')
    await assert_job_count(client=client, n=3)

    await unschedule_action(client=client, action_name='foo1') # two schedulers involved
    await assert_job_count(client=client, n=1)
    await get_action(client=client, action_name='foo1')
    await get_action(client=client, action_name='foo2')
    await get_scheduler(client=client, scheduler_name='bar1')
    await get_scheduler(client=client, scheduler_name='bar2')

@pytest.mark.asyncio
async def test_reschedule_action_1(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """ schedule action, run it, change it, re-run it """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    action1 = FileHeartbeat(file=str(tmp_path / 'output1.txt'))
    action2 = FileHeartbeat(file=str(tmp_path / 'output2.txt'))
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo', action=action1)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo', scheduler_name='bar')
    await assert_job_count(client=client, n=1)

    await run_and_stop_jobs(client=client, pause=4)
    lines = None
    with open(action1.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1

    await set_action(client=client, action_name='foo', action=action2)
    await reschedule_action(client=client, action_name='foo')
    await assert_job_count(client=client, n=1)

    await run_and_stop_jobs(client=client, pause=4)
    lines = None
    with open(action2.file, 'r') as fid:
        lines = fid.readlines()
    assert lines is not None and isinstance(lines, list) and len(lines) >= 1
        
@pytest.mark.asyncio
async def test_unschedule_scheduler(startup_and_shutdown_uvicorn, host, port, tmp_path):
    """ unschedule a scheduler. """
    client = ClientAsync(host=host, port=port)
    await reset_dispatcher(client, str(tmp_path))

    action1 = FileHeartbeat(file=str(tmp_path / 'output1.txt'))
    action2 = FileHeartbeat(file=str(tmp_path / 'output2.txt'))
    scheduler = TimelyScheduler(interval=1)

    await add_action(client=client, action_name='foo1', action=action1)
    await add_action(client=client, action_name='foo2', action=action2)
    await add_scheduler(client=client, scheduler_name='bar', scheduler=scheduler)
    await schedule_action(client=client, action_name='foo1', scheduler_name='bar')
    await schedule_action(client=client, action_name='foo2', scheduler_name='bar')
    await assert_job_count(client=client, n=2)

    await unschedule_scheduler(client=client, scheduler_name='bar')
    await assert_job_count(client=client, n=0)
    await get_action(client=client, action_name='foo1')
    await get_action(client=client, action_name='foo2')
    await get_scheduler(client=client, scheduler_name='bar')

# helpers

async def get_action(client:ClientAsync, action_name:str):
    """ make sure action exists and resolves properly """
    retrieved_action = await client.get_action(action_name=action_name)
    assert isinstance(retrieved_action, Action), str(type(retrieved_action))
    return retrieved_action

async def add_action(client:ClientAsync, action_name:str, action:Action):
    """ add an action and confirm """
    response = await client.add_action(action_name=action_name, action=action)
    assert response.status_code == 200, f"failed to put action ({action_name})"
    retrieved_action = await client.get_action(action_name=action_name)
    assert isinstance(retrieved_action, Action), str(type(retrieved_action))

async def set_action(client:ClientAsync, action_name:str, action:Action):
    """ set an action and confirm """
    response = await client.set_action(action_name=action_name, action=action)
    assert response.status_code == 200, f"failed to put action ({action_name})"
    retrieved_action = await client.get_action(action_name=action_name)
    assert isinstance(retrieved_action, Action), str(type(retrieved_action))

async def unschedule_action(client:ClientAsync, action_name:str):
    response = await client.unschedule_action(action_name=action_name)
    assert response.status_code == 200, f"failed to unschedule action ({action_name})"

async def reschedule_action(client:ClientAsync, action_name:str):
    response = await client.reschedule_action(action_name=action_name)
    assert response.status_code == 200, f"failed to reschedule action ({action_name}) with response({response.json()})"

async def get_scheduler(client:ClientAsync, scheduler_name:str):
    """ make sure scheduler exists and resolves properly """
    retrieved_scheduler = await client.get_scheduler(scheduler_name=scheduler_name)
    assert isinstance(retrieved_scheduler, Scheduler), str(type(retrieved_scheduler))
    return retrieved_scheduler

async def add_scheduler(client:ClientAsync, scheduler_name:str, scheduler:Scheduler):
    """ add a scheduler and confirm """
    response = await client.add_scheduler(scheduler_name=scheduler_name, scheduler=scheduler)
    assert response.status_code == 200, f"failed to put scheduler ({scheduler_name})"
    retrieved_scheduler = await client.get_scheduler(scheduler_name=scheduler_name)
    assert isinstance(retrieved_scheduler, Scheduler), str(type(retrieved_scheduler))

async def set_scheduler(client:ClientAsync, scheduler_name:str, scheduler:Scheduler):
    """ add a scheduler and confirm """
    response = await client.set_scheduler(scheduler_name=scheduler_name, scheduler=scheduler)
    assert response.status_code == 200, f"failed to put scheduler ({scheduler_name})"
    retrieved_scheduler = await client.get_scheduler(scheduler_name=scheduler_name)
    assert isinstance(retrieved_scheduler, Scheduler), str(type(retrieved_scheduler))

async def unschedule_scheduler(client:ClientAsync, scheduler_name:str):
    response = await client.unschedule_scheduler(scheduler_name=scheduler_name)
    assert response.status_code == 200, f"failed to unschedule scheduler ({scheduler_name})"

async def schedule_action(client:ClientAsync, scheduler_name:str, action_name:str):
    """ schedule an action """
    response = await client.schedule_action(scheduler_name=scheduler_name, action_name=action_name)
    assert response.status_code == 200, f"failed to schedule action ({action_name}) using scheduler ({scheduler_name})"

async def reset_dispatcher(client:ClientAsync, tmp_dir:str):
    """
    usage: reset_dispatcher(client, str(tmp_path))
    """
    # set saved_dir to fixture's tmp_path (see usage)
    saved_dir = FilePathe(path=tmp_dir)
    response = await client.set_saved_dir(saved_dir=saved_dir)
    assert response.status_code == 200, 'failed to set saved_dir'
    saved_saved_dir = await client.get_saved_dir()
    assert saved_saved_dir == saved_dir

    # empty the dispatcher and stop the jobs if they're running
    # for uvicorn tests to run in a bunch, these objects need to be 'reset' since
    # evidently the app is shared during the running of the tests
    response = await client.clear_dispatcher()
    assert response.status_code == 200, 'failed to clear dispatcher'
    response = await client.stop_jobs() # likely already stopped

async def assert_job_count(client:ClientAsync, n:int):
    response = await client.job_count()
    assert response.status_code == 200, 'failed to retrieve job count'
    job_count = response.json()['job_count']
    assert job_count == n, f"expected a job count of ({n})"

async def run_and_stop_jobs(client:ClientAsync, pause:int):
    response = await client.run_jobs()
    assert response.status_code == 200, 'failed to start jobs'
    time.sleep(pause)
    response = await client.stop_jobs()
    assert response.status_code == 200, 'failed to stop jobs'


from pydantic import BaseModel
from httpx import AsyncClient
from whendo.core.action import Action
from whendo.core.scheduler import Scheduler
from whendo.core.resolver import resolve_action, resolve_scheduler, resolve_file_pathe
from whendo.core.util import FilePathe
from whendo.core.dispatcher import Dispatcher

class ClientAsync(BaseModel):
    """
    This class is very similar to Client, except that the methods
    that typically return BaseModel objects will return supplied
    dictionaries if resolution fails. 
    """
    host:str='127.0.0.1'
    port:int=8000

    # /dispatcher
    async def load_dispatcher(self):
        return Dispatcher.resolve(await self.get_as_json(f"/dispatcher/load"))
    async def save_dispatcher(self):
        return await self.get("/dispatcher/save")
    async def clear_dispatcher(self):
        return await self.get("/dispatcher/clear")
    async def load_dispatcher_from_name(self, name:str):
        return Dispatcher.resolve(await self.get_as_json(f"/dispatcher/load_from_name/{name}"))
    async def save_data_to_name(self, name:str):
        return await self.get(f"/dispatcher/save_to_name/{name}")
    async def get_saved_dir(self):
        return resolve_file_pathe(await self.get_as_json("/dispatcher/saved_dir"), check_for_found_class=False)
    async def set_saved_dir(self, saved_dir:FilePathe):
        return await self.put("/dispatcher/saved_dir", saved_dir)

    # /actions
    async def get_action(self, action_name:str):
        return resolve_action(await self.get_as_json(f"/actions/{action_name}"), check_for_found_class=False)
    async def add_action(self, action_name:str, action:Action):
        return await self.post(f"/actions/{action_name}", action)
    async def set_action(self, action_name:str, action:Action):
        return await self.put(f"/actions/{action_name}", action)
    async def delete_action(self, action_name:str):
        return await self.delete(f"/actions/{action_name}")
    async def execute_action(self, action_name:str):
        return await self.get(f"/actions/{action_name}/execute")
    async def reschedule_action(self, action_name:str):
        return await self.get(f"/actions/{action_name}/reschedule")
    async def unschedule_action(self, action_name:str):
        return await self.get(f"/actions/{action_name}/unschedule")

    # /schedulers
    async def schedule_action(self, scheduler_name:str, action_name:str):
        return await self.get(f"/schedulers/{scheduler_name}/actions/{action_name}")
    async def get_scheduler(self, scheduler_name:str):
        return resolve_scheduler(await self.get_as_json(f"/schedulers/{scheduler_name}"), check_for_found_class=False)
    async def add_scheduler(self, scheduler_name:str, scheduler:Scheduler):
        return await self.post(f"/schedulers/{scheduler_name}", scheduler)
    async def set_scheduler(self, scheduler_name:str, scheduler:Scheduler):
        return await self.put(f"/schedulers/{scheduler_name}", scheduler)
    async def delete_scheduler(self, scheduler_name:str):
        return await self.delete(f"/schedulers/{scheduler_name}")
    async def unschedule_scheduler(self, scheduler_name:str):
        return await self.get(f"/schedulers/{scheduler_name}/unschedule")
    async def reschedule_all_schedulers(self):
        return await self.get(f"/schedulers/reschedule_all")
    async def execute_scheduler_actions(self, scheduler_name:str):
        return await self.get(f"/schedulers/{scheduler_name}/execute")

    # /jobs
    async def run_jobs(self):
        return await self.get(f"/jobs/run")
    async def stop_jobs(self):
        return await self.get(f"/jobs/stop")
    async def jobs_are_running(self):
        return await self.get(f"/jobs/are_running")
    async def job_count(self):
        return await self.get(f"/jobs/count")
    async def clear_jobs(self):
        return await self.get(f"/jobs/clear")

    # verbs
    async def get(self, path:str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.get(url=path)
    async def get_as_json(self, path:str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            response = await ac.get(url=path)
            return response.json()
    async def put(self, path:str, data:BaseModel):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.put(url=path, data=data.json())
    async def post(self, path:str, data:BaseModel):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.post(url=path, data=data.json())
    async def delete(self, path:str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.delete(url=path)

    def base_url(self):
        return f"http://{self.host}:{self.port}"


from datetime import datetime, timedelta
from pydantic import BaseModel
from typing import List
import whendo.core.util as util

def test_now_1():
    now1 = datetime.now()
    now2 = util.Now.dt()
    elapsed = now2 - now1
    assert elapsed < timedelta(microseconds=100.0)

def test_now_2():
    now = datetime.now()
    now2 = util.Now.t()
    date2 = datetime(year=now.year, month=now.month, day=now.day, hour=now2.hour, minute=now2.minute, second=now2.second, microsecond=now2.microsecond)
    elapsed = date2 - now
    assert elapsed < timedelta(microseconds=100.0)

def test_now_3():
    now1 = datetime.now().time()
    now1_str = now1.strftime('%H:%M:%S')
    now2 = util.Now.s()
    assert now1_str in now2

def test_all_visible_subclasses():
    class A:
        pass
    class B(A):
        pass
    class C(A):
        pass
    class D(B):
        pass
    x = util.all_visible_subclasses(A)
    y = {B,C,D}
    assert x == y

def test_key_strings_from_class():
    class A(BaseModel):
        a:str
        b:str
    x = {'a', 'b'}
    y = util.key_strings_from_class(A)
    assert x == y

def test_key_strings_from_dict():
    x = {'a', 'b'}
    y = util.key_strings_from_dict({'a':1, 'b':2})
    assert x == y

def test_resolve_instance():
    """
    Top > A, B, D
    D > C
    """
    class Top(BaseModel):
        pass
    class C(Top):
        c:str
    class D(C):
        d:str
        d_instance:Top
    class B(Top):
        b_instance:Top
        b_list:List[Top]
    class A(Top):
        a_list:List[Top]

    dictionary0 = {'c':'why'}
    dictionary1 = { # D
                        'c':'dc',
                        'd':'abc',
                        'd_instance': { # C
                            'c':'mmm',
                        }
                    }
    dictionary2 = { # A
        'a_list': [
            { # B
                'b_instance': { # C
                    'c':'xxx'
                    },
                'b_list': [
                    {'c':'yyy'}, # C
                    {'c':'zzz'}, # C
                    dictionary1
                    ]
            },
            { # D
                'c':'people',
                'd':'who',
                'd_instance': dictionary0
            }
        ]
    }
    instance0 = util.resolve_instance(Top, dictionary0)
    instance1 = util.resolve_instance(Top, dictionary1)
    instance2 = util.resolve_instance(Top, dictionary2)
    assert isinstance(instance0, C)
    assert isinstance(instance1, D)
    assert isinstance(instance2, A)
    assert isinstance(instance2.a_list[0].b_instance, C)
    assert instance1 == instance2.a_list[0].b_list[2]
    assert instance0 == instance2.a_list[1].d_instance

def test_filepathe():
    path = util.FilePathe(path = 'a/b/c')
    new_path = util.FilePathe(**(path.dict()))
    assert path.path == new_path.path, 'path reconstruction failed'

def test_dirs(tmp_path):
    def assure_writes_and_reads(label, thunk):
        lines = None
        file = thunk(root_path=tmp_path) + f"saved_{label}.txt"
        with open(file, 'w') as fid:
            fid.write('blee blee blee\n')
        with open(file, 'r') as fid:
            lines = fid.readlines()
        assert lines is not None and isinstance(lines, list) and len(lines) >= 1
        return file
    file1 = assure_writes_and_reads('foo', util.Dirs.saved_dir)
    file2 = assure_writes_and_reads('foo', util.Dirs.output_dir)
    file3 = assure_writes_and_reads('foo', util.Dirs.log_dir)
    assert file1 != file2 and file2 != file3 and file3 != file1

def test_shared_rw_1():
    shared = util.SharedRWs.get('foo')
    def funk(dictionary:dict):
        dictionary['a'] = 1
        return 1
    result = shared.apply(funk)
    assert result == 1
    assert isinstance(shared.data_copy(), dict)

def test_shared_rw_2():
    """
    Test transactional nature with a funky failure.
    """
    shared = util.SharedRWs.get('foo')
    def funk1(dictionary:dict):
        dictionary['a'] = 1
        return 1
    result = shared.apply(funk1)
    def funk2(dictionary:dict):
        dictionary['a'] = 2
        raise Exception('oops')
    try:
        shared.apply(funk2)
    except:
        pass
    assert shared.data_copy()['a'] == 1


def test_shared_ro_1():
    dictionary = {'a':1}
    shared = util.SharedROs.get('foo', dictionary)
    dictionary['a'] = 3
    assert shared.data_copy()['a'] == 1



import whendo.core.actions.logic_action as logic_x
from whendo.core.action import Action
import pytest

def negate(action:Action):
    return logic_x.NotAction(operand=action)

def is_exception(result):
    return isinstance(result, Exception)

def computes_exception(thunk):
    try:
        result = thunk()
    except Exception as exception:
        result = exception
    return isinstance(result, Exception)

# =================== tests ==================?

def test_success_1():
    assert not computes_exception(logic_x.SuccessAction().execute)

def test_exception_1():
    assert computes_exception(logic_x.ExceptionAction().execute)

def test_not_1():

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            raise Exception()

    action1 = Action1()
    action2 = negate(action1)
    action3 = negate(action2)
    assert computes_exception(action1.execute)
    assert not computes_exception(action2.execute)
    assert computes_exception(action3.execute)

def test_not_2():
    
    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            return True

    action1 = Action1()
    action2 = negate(action1)
    action3 = negate(action2)
    assert not computes_exception(action1.execute)
    assert computes_exception(action2.execute)
    assert not computes_exception(action3.execute)

def test_list_action_all_1():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
    
    list_action = logic_x.ListAction(op_mode=logic_x.ListOpMode.ALL, action_list=[Action1(), Action2(), Action3(), Action4()])
    result = list_action.execute()
    assert dictionary['value'] == Action4

def test_list_action_or_1():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__

    list_action = logic_x.ListAction(op_mode=logic_x.ListOpMode.OR, action_list=[negate(Action1()), Action2(), Action3(), Action4()])
    result = list_action.execute()
    assert dictionary['value'] == Action2

def test_list_action_and_1():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__

    list_action = logic_x.ListAction(op_mode=logic_x.ListOpMode.AND, action_list=[Action1(), Action2(), Action3(), Action4()])
    result = list_action.execute()
    assert dictionary['value'] == Action4

def test_list_action_and_2():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__

    list_action = logic_x.ListAction(
        op_mode=logic_x.ListOpMode.AND,
        action_list=[negate(Action1()), negate(Action2()), negate(Action3()), negate(Action4())])
    result = list_action.execute()
    assert dictionary['value'] == Action1

def test_list_action_and_3():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            return Exception()
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            return Exception()
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            return Exception()
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            return Exception()

    list_action = logic_x.ListAction(
        op_mode=logic_x.ListOpMode.AND,
        action_list=[Action1(), Action2(), Action3(), Action4()],
        exception_on_no_success=True
        )
    assert computes_exception(list_action.execute)

def test_if_else_action_and_1():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True

    list_action = logic_x.IfElseAction(
        op_mode=logic_x.ListOpMode.AND,
        test_action=Action1(),
        else_action=Action2(),
        if_actions=[Action3(), Action4()],
        exception_on_no_success=False
        )
    result = list_action.execute()
    assert dictionary['value'] == Action4

def test_if_else_action_or_2():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True

    list_action = logic_x.IfElseAction(
        op_mode=logic_x.ListOpMode.OR,
        test_action=Action1(),
        else_action=Action2(),
        if_actions=[Action3(), Action4()],
        exception_on_no_success=False
        )
    result = list_action.execute()
    assert dictionary['value'] == Action3

def test_if_else_action_else_1():
    dictionary = {'value':None}

    class Action1(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            return Exception()
        
    class Action2(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action3(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True
        
    class Action4(Action):
        def execute(self, tag:str=None, scheduler_info:dict=None):
            dictionary['value'] = self.__class__
            return True

    list_action = logic_x.IfElseAction(
        op_mode=logic_x.ListOpMode.OR,
        test_action=Action1(),
        else_action=Action2(),
        if_actions=[Action3(), Action4()],
        exception_on_no_success=False
        )
    result = list_action.execute()
    assert dictionary['value'] == Action2



from fastapi import APIRouter, status, Depends
from whendo.api.shared import return_success, raised_exception, get_dispatcher
from whendo.core.resolver import resolve_action

router = APIRouter(
    prefix='',
    tags=['Actions']
)

@router.get('/actions/{action_name}', status_code=status.HTTP_200_OK)
def get_action(action_name:str):
    try:
        action = get_dispatcher(router).get_action(action_name=action_name)
        return return_success(action)
    except Exception as e:
        raise raised_exception(f"failed to retrieve the action ({action_name})", e)

@router.post('/actions/{action_name}', status_code=status.HTTP_200_OK)
def add_action(action_name:str, action=Depends(resolve_action)):
    try:
        assert action, f"couldn't resolve class for action ({action_name})"
        get_dispatcher(router).add_action(action_name=action_name, action=action)
        return return_success(f"action ({action_name}) was successfully added")
    except Exception as e:
        raise raised_exception(f"failed to add action ({action_name})", e)

@router.put('/actions/{action_name}', status_code=status.HTTP_200_OK)
def set_action(action_name:str, action=Depends(resolve_action)):
    try:
        assert action, f"couldn't resolve class for action ({action_name})"
        get_dispatcher(router).set_action(action_name=action_name, action=action)
        return return_success(f"action ({action_name}) was successfully updated")
    except Exception as e:
        raise raised_exception(f"failed to update action ({action_name})", e)

@router.delete('/actions/{action_name}', status_code=status.HTTP_200_OK)
def delete_action(action_name:str):
    try:
        get_dispatcher(router).delete_action(action_name=action_name)
        return return_success(f"action ({action_name}) was successfully deleted")
    except Exception as e:
        raise raised_exception(f"failed to delete action ({action_name})", e)

@router.get('/actions/{action_name}/execute', status_code=status.HTTP_200_OK)
def execute_action(action_name:str):
    """
    Two potential types of exceptions below. One resulting from the actual execute_action call and
    the other returned from the execution of the action itself.
    """
    exception = None # the action's exception if the action generated an exception
    try:
        result = get_dispatcher(router).execute_action(action_name=action_name)
        if not isinstance(result, Exception):
            return return_success({'msg': f"action ({action_name}) was successfully executed", 'result':result})
        else:
            exception = raised_exception(f"failed to execute action ({action_name})", result)
    except Exception as e: # from execute_action
        raise raised_exception(f"failed to execute action ({action_name})", e)
    if exception is not None: # from the action
        raise exception

@router.get('/actions/{action_name}/unschedule', status_code=status.HTTP_200_OK)
def unschedule_action(action_name:str):
    try:
        get_dispatcher(router).unschedule_action(action_name=action_name)
        return return_success(f"action ({action_name}) was successfully unscheduled")
    except Exception as e:
        raise raised_exception(f"failed to unschedule action ({action_name})", e)

@router.get('/actions/{action_name}/reschedule', status_code=status.HTTP_200_OK)
def reschedule_action(action_name:str):
    try:
        get_dispatcher(router).reschedule_action(action_name=action_name)
        return return_success(f"action ({action_name}) was successfully rescheduled")
    except Exception as e:
        raise raised_exception(f"failed to reschedule action ({action_name})", e)


from pydantic import BaseModel
from whendo.sdk.client import Client
    
class Workbench(BaseModel):
    clients:dict={}
    def add_client(self, name:str, client:Client):
        self.clients[name] = client
    def get_client(self, name:str):
        return self.clients.get(name, None)
    def get_clients(self):
        return list(self.clients.values())
    def count(self):
        return len(self.clients)


import requests
from typing import Optional, Dict, Any
from whendo.core.action import Action

class SendPayload(Action):
    """
    This class sends a payload dictionary to a url.
    """
    url:str
    payload:Optional[Dict[str, Any]]=None

    def execute(self, tag:str=None, scheduler_info:dict=None):
        response = requests.get(self.url, self.payload)
        if response.status_code != requests.codes.ok:
            raise Exception(response)
        return response.json()


"""
This code in this module is used in FastAPI and APIRouter code.
"""
from fastapi import status, HTTPException, APIRouter
from whendo.core.dispatcher import Dispatcher
from whendo.core.util import Now

def return_success(obj:object):
    """
    for successful returns
    """
    return obj

def raised_exception(text:str, exception:Exception):
    """
    for exception reporting
    """
    status_code = status.HTTP_400_BAD_REQUEST
    detail = {'outcome':text, 'exception': str(exception), 'time':Now.s()}
    return HTTPException(status_code=status_code, detail=detail)

# these functions enabling the passing down of singletons to routers from the main app
def get_dispatcher(router:APIRouter):
    return router.__dict__['_dispatcher']

def set_dispatcher(router:APIRouter, dispatcher:Dispatcher):
    router.__dict__['_dispatcher'] = dispatcher
    return router


"""
This class implements a single process-based system for defining and scheduling actions local
to the python runtime (api server, testing harness). The class implements all of the potential
endpoints of a restful api and supports testing outside of the restful api implementation.

Instances of this class contain Schedulers and Actions, which can at any point be submitted to and removed from the
job scheduling mechanism of the schedule library (refer to the 'continuous' module).
"""
from pydantic import BaseModel, PrivateAttr
from threading import RLock
import pickle
import json
import os
import logging
from typing import Optional, Dict, Any
from whendo.core.util import PP, Dirs
from whendo.core.action import Action
from whendo.core.scheduler import Scheduler
from whendo.core.continuous import Continuous
from whendo.core.resolver import resolve_action, resolve_scheduler

logger = logging.getLogger()

class Dispatcher(BaseModel):
    """
    Serializations of this class are stored in the local file system. When a runtime starts
    up, Data loads the last saved version.
    """
    actions: dict={}
    schedulers: dict={}
    schedulers_actions: dict={}
    saved_dir: Optional[str]=None

    # not treated as a model attr
    _continuous: Continuous = PrivateAttr(default_factory = Continuous.get)

    def get_continuous(self):
        return self._continuous
    def set_continuous(self, continuous:Continuous):
        self._continuous = continuous

    def run_jobs(self):
        self._continuous.run_continuously()
    def stop_jobs(self):
        self._continuous.stop_running_continuously()
    def jobs_are_running(self):
        return self._continuous.is_running()
    def job_count(self):
        return self._continuous.job_count()
    def clear_jobs(self):
        self._continuous.clear()

    def get_actions(self):
        with Lok.lock:
            return self.actions
    def get_schedulers(self):
        with Lok.lock:
            return self.schedulers
    def get_schedulers_actions(self):
        with Lok.lock:
            return self.schedulers_actions

    def initialize(self):
        Lok.reset()
    def pprint(self):
        PP.pprint(self.dict())
    
    def load_current(self):
        return self.load_from_name('current')
    def save_current(self):
        with Lok.lock:
            self.save_to_name('current')

    def load_from_name(self, name:str):
        with Lok.lock:
            assert self.saved_dir, 'saved_dir must be set'
            with open(self.saved_dir+name+'.json', 'r') as infile:
                json_string = json.load(infile)
                dictionary = json.loads(json_string)
                return Dispatcher.resolve(dictionary)
    def save_to_name(self, name:str):
        with Lok.lock:
            assert self.saved_dir, 'saved_dir must be set'
            with open(self.saved_dir+name+'.json', 'w') as outfile:
                json.dump(self.json(), outfile, indent=2)

    def get_saved_dir(self):
        return self.saved_dir
    def set_saved_dir(self, saved_dir:str):
        with Lok.lock:
            if not os.path.exists(saved_dir):
                os.makedirs(saved_dir)
            self.saved_dir = saved_dir
            self.save_current()

    def clear_all(self):
        with Lok.lock:
            schedulers_copy = self.schedulers.copy()
            for scheduler_name in schedulers_copy:
                self.delete_scheduler(scheduler_name)
            actions_copy = self.actions.copy()
            for action_name in actions_copy:
                self.delete_action(action_name)
            self.schedulers_actions.clear()
            self.save_current()

    # actions
    def get_action(self, action_name:str):
        with Lok.lock:
            # assert action_name in self.actions, f"action ({action_name}) does not exist"
            return self.actions.get(action_name, None)
    def add_action(self, action_name:str, action:Action):
        with Lok.lock:
            assert not action_name in self.actions, f"action ({action_name}) already exists"
            self.actions[action_name] = action
            self.save_current()
    def set_action(self, action_name:str, action:Action):
        with Lok.lock:
            assert action_name in self.actions, f"action ({action_name}) does not exist"
            self.actions[action_name] = action
            self.reschedule_action(action_name)
            self.save_current()
    def delete_action(self, action_name:str):
        with Lok.lock:
            assert action_name in self.actions, f"action ({action_name}) does not exist"
            for scheduler_name in self.schedulers:
                self.unschedule_action(action_name)
                action_names = self.schedulers_actions[scheduler_name]
                if action_name in action_names: action_names.remove(action_name)
            self.actions.pop(action_name, None)
            self.save_current()
    def execute_action(self, action_name:str):
        with Lok.lock:
            assert action_name in self.actions, f"action ({action_name}) does not exist"
            return self.get_action(action_name).execute()

    # schedulers
    def get_scheduler(self, scheduler_name:str):
        with Lok.lock:
            return self.schedulers.get(scheduler_name, None)
    def add_scheduler(self, scheduler_name:str, scheduler:Scheduler):
        with Lok.lock:
            assert not scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) already exists"
            self.schedulers[scheduler_name] = scheduler
            self.schedulers_actions[scheduler_name] = []
            self.save_current()
    def set_scheduler(self, scheduler_name:str, scheduler:Scheduler):
        with Lok.lock:
            assert scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) does not exist"
            self.schedulers[scheduler_name] = scheduler
            self.reschedule_scheduler(scheduler_name)
            self.save_current()
    def delete_scheduler(self, scheduler_name:str):
        with Lok.lock:
            assert scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) does not exist"
            self.unschedule_scheduler(scheduler_name)
            self.schedulers.pop(scheduler_name)
            self.schedulers_actions[scheduler_name].clear() # pro-actively clean up. less work for GC.
            self.schedulers_actions.pop(scheduler_name)
            self.save_current()
    def execute_scheduler_actions(self, scheduler_name:str):
        with Lok.lock:
            assert scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) does not exist"
            result = []
            for action_name in self.schedulers_actions.get(scheduler_name, []):
                result.append(self.get_action(action_name).execute())
            return result

    # scheduling
    def schedule_action(self, scheduler_name:str, action_name:str):
        with Lok.lock:
            assert scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) does not exist"
            assert action_name in self.actions, f"action ({action_name}) does not exist"
            assert not action_name in self.schedulers_actions[scheduler_name]
            scheduler = self.get_scheduler(scheduler_name)
            action = self.actions.get(action_name)
            tag = self.scheduler_tag(scheduler_name, action_name)
            scheduler.schedule_action(tag, action, self._continuous)
            self.schedulers_actions[scheduler_name].append(action_name)
            self.save_current()
    def unschedule_action(self, action_name:str):
        with Lok.lock:
            assert action_name in self.actions, f"action ({action_name}) does not exist"
            for scheduler_name in self.schedulers_actions:
                if action_name in self.schedulers_actions[scheduler_name]:
                    tag = self.scheduler_tag(scheduler_name, action_name)
                    self._continuous.clear(tag)
                    action_names = self.schedulers_actions[scheduler_name]
                    if action_name in action_names: action_names.remove(action_name)
            self.save_current()
    def reschedule_action(self, action_name:str):
        with Lok.lock:
            assert action_name in self.actions, f"action ({action_name}) does not exist"
            for scheduler_name in self.schedulers_actions:
                if action_name in self.schedulers_actions[scheduler_name]:
                    tag = self.scheduler_tag(scheduler_name, action_name)
                    self._continuous.clear(tag)
                    scheduler = self.get_scheduler(scheduler_name)
                    action = self.actions[action_name]
                    scheduler.schedule_action(tag, action, self._continuous)
    def unschedule_scheduler(self, scheduler_name:str):
        with Lok.lock:
            assert scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) does not exist"
            for action_name in self.schedulers_actions[scheduler_name]:
                tag = self.scheduler_tag(scheduler_name, action_name)
                self._continuous.clear(tag)
            self.schedulers_actions[scheduler_name].clear()
            self.save_current()
    def reschedule_scheduler(self, scheduler_name:str):
        with Lok.lock:
            assert scheduler_name in self.schedulers, f"scheduler ({scheduler_name}) does not exist"
            scheduler_actions = self.schedulers_actions[scheduler_name]
            for action_name in scheduler_actions:
                tag = self.scheduler_tag(scheduler_name, action_name)
                self._continuous.clear(tag)
                scheduler = self.get_scheduler(scheduler_name)
                action = self.actions[action_name]
                scheduler.schedule_action(tag, action, self._continuous)
    def reschedule_all_schedulers(self):
        with Lok.lock:
            for scheduler_name in self.schedulers_actions:
                for action_name in self.schedulers_actions[scheduler_name]:
                    tag = self.scheduler_tag(scheduler_name, action_name)
                    self._continuous.clear(tag)
                    scheduler = self.get_scheduler(scheduler_name)
                    action = self.actions[action_name]
                    scheduler.schedule_action(tag, action, self._continuous)
    def get_scheduled_action_count(self):
        # returns the total number of actions in the schedulers_actions dictionary
        # should be equal to the number of jobs in related Continuous instance
        with Lok.lock:
            return sum(len(action_array) for action_array in self.schedulers_actions.values())
  
    # utility
    def scheduler_tag(self, scheduler_name:str, action_name:str):
        # computes job tag for scheduler and action
        return f"{scheduler_name}:{action_name}"
        
    @classmethod
    def resolve(cls, dictionary:dict={}):
        """
        converts dictionary to an Data instance.
        """
        if len(dictionary) == 0:
            return Dispatcher(saved_dir=Dirs.saved_dir)
        else:
            saved_dir = dictionary['saved_dir']
            actions = dictionary['actions']
            schedulers = dictionary['schedulers']
            schedulers_actions = dictionary['schedulers_actions']
            # replace key's value for each key...
            for action_name in actions:
                actions[action_name] = resolve_action(actions[action_name])
            for scheduler_name in schedulers:
                schedulers[scheduler_name] = resolve_scheduler(schedulers[scheduler_name])
            return Dispatcher(
                saved_dir=saved_dir,
                actions=actions,
                schedulers=schedulers,
                schedulers_actions=schedulers_actions)

# ------------------
"""
usage:
  with Lok.lock:
    # critical section
note:
  singleton
"""
class Lok:
    lock = RLock()
    @classmethod
    def reset(cls):
        cls.lock = RLock()

class DispatcherSingleton:
    # this call returns the standard non-test Data singleton
    dispatcher = None
    @classmethod
    def get(cls):
        if not cls.dispatcher:
            cls.dispatcher = Dispatcher(saved_dir=Dirs.saved_dir())
            try:    
                cls.dispatcher = cls.dispatcher.load_current()
            except Exception as exception:
                logger.error("error loading dispatcher from disk", exception)
            cls.dispatcher.set_continuous(Continuous.get())
            cls.dispatcher.initialize()
            cls.dispatcher.reschedule_all_schedulers()
        return cls.dispatcher


from datetime import time
from pydantic import BaseModel
from typing import Dict, Optional
from collections.abc import Callable
from threading import RLock
import logging
import random
from whendo.core.util import TimeUnit, Now, object_info
from whendo.core.action import Action
from whendo.core.continuous import Continuous

logger = logging.getLogger(__name__)

class Count:
    """
    This counter is useful for distiguishing between different sets of log entries as well as viewing a count for each tag.
    """
    good_count:Dict[str,int] = {}
    bad_count:Dict[str,int]  = {}
    @classmethod
    def good_up(cls, tag):
        if tag not in cls.good_count:
            cls.good_count[tag] = 0
        cls.good_count[tag] = 1 + cls.good_count[tag]
    @classmethod
    def bad_up(cls, tag):
        if tag not in cls.bad_count:
            cls.bad_count[tag] = 0
        cls.bad_count[tag] = 1 + cls.bad_count[tag]
    @classmethod
    def good(cls, tag):
        return cls.good_count.get(tag, 0)
    @classmethod
    def bad(cls, tag):
        return cls.bad_count.get(tag, 0)

class DoNothing:
    """
    Returned by the do_nothing thunk. The exception handling
    code needs to know when a result was produced by a do_nothing thunk.
    In the abasence of an explicitly returned result, an Action will return
    None. 
    """
    result = {"outcome":"DoNothing"}

class Scheduler(BaseModel):
    """
    An instance of the Scheduler class schedules Action instances using the Python 'schedule' library. This
    library schedules and runs 'jobs'. A submitted job can be any zero-arg Callable.
    
        https://schedule.readthedocs.io/en/stable/
    
    All Schedulers can limit the time during the day that their jobs run by setting the 'start' and 'stop'
    fields to instances of the 'datetime.datetime.time' class.
    
    examples:
        start=time(8,0,0) stop=time(18,0,0) limits execution to the time between 8:00 and 18:00
        start=time(18,0,0) stop=time(8,0,0) limits execution to the time between 18:00 and 8:00
    
    note:
        1. This (start, stop) in-period feature is implemented outside of the workings of the schedule library since schedule's schedule build 
           mechanism does not support it. So it's done using a less than optimal approach that alters the execution
           behavior within a scheduled job. Outside of the specified period, a job will be run. It just won't do anything. [see do_nothing below]
    """
    start:Optional[time] = None
    stop:Optional[time] = None
    
    def schedule_action(self, tag:str, action:Action, continuous:Continuous):
        pass
  
    def during_period(self, tag:str, action:Action):
        """
        is_in_period_wrapper below takes (start) and (stop) and returns a 1-arg function that compares (start)
        and (stop) with a supplied time (now).

        if start<stop, then the active period is defined by
            start<now AND now<stop
            e.g. - 6:00 => 18:00
        if start>stop, then the period period is defined by
            start<now OR now<stop
            e.g. - 18:00 => 6:00
        
        This is done so that is_in_period is freshly evaluated at the times when the schedule library runs the job
        (that is meant to invoke the Callable).
        """
        if (self.start is not None) and (self.stop is not None):
            is_in_period_wrapper = lambda start, stop: (lambda now: start < now and now < stop if (start < stop) else start < now or now < stop)
            is_in_period = is_in_period_wrapper(self.start, self.stop)
            do_nothing = lambda tag, scheduler_info: DoNothing.result
            return self.wrap(lambda: (action.execute if is_in_period(Now.t()) else do_nothing)(tag=tag, scheduler_info=self.info()), tag=tag, action_json=action.json(), scheduler_json=self.json())
        else:
            return self.wrap(lambda: action.execute(tag=tag, scheduler_info=self.info()), tag=tag, action_json=action.json(), scheduler_json=self.json())
    
    def wrap(self, thunk:Callable, tag:str, action_json:str, scheduler_json:str):
        """
        wraps the execution inside a try block;
        allows for logging and handling of raised exceptions
        """
        def execute():
            try:
                result = thunk()
                if result is DoNothing.result:
                    return result
                else:
                    Count.good_up(tag)
                    logger.info(f"({Count.good(tag):09}) tag={tag}")
                    logger.info(f"({Count.good(tag):09}) scheduler={scheduler_json}")
                    logger.info(f"({Count.good(tag):09}) action={action_json}")
            except Exception as exception:
                result = exception
                Count.bad_up(tag)
                logger.exception(f"({Count.bad(tag):09}) tag={tag}", exc_info=exception)
                logger.exception(f"({Count.bad(tag):09}) scheduler={scheduler_json}", exc_info=exception)
                logger.exception(f"({Count.bad(tag):09}) action={action_json}", exc_info=exception)
            return result
        return execute

    def info(self):
        return object_info(self)
    
class TimelyScheduler(Scheduler):
    """
    This scheduler executes Actions every [interval] days, hours, minutes, seconds at a
    time within the time unit (hour, minute, second).
    
    The time unit is determined by the values of the hour, minute, and second
    fields.

    If hour is specified, the time unit is set to day. [17:00:00]
    If minute is specified, the time unit is set to hour. [:15:00]
    If second is specified, the time unit is set to minute. [:15]
    Otherwise the time unit is set to second.

    For example, for hour=None, minute=15, second=None, interval=2, the system will
    execute the action every 2 hours at 15 minutes past the hour.
    """
    interval: int
    hour:Optional[int] = None
    minute:Optional[int] = None
    second:Optional[int] = None

    def schedule_action(self, tag:str, action:Action, continuous:Continuous):
        callable = self.during_period(tag=tag, action=action)
        continuous.schedule_timely_callable(
            tag=tag,
            callable=callable,
            interval=self.interval,
            hour=self.hour, minute=self.minute, second=self.second
        )

class RandomlyScheduler(Scheduler):
    """
    This scheduler randomly executes Actions every [low] to [high] days, hours, minutes, or seconds.
    
    The time unit is supplied at instance creation.

    For example, for low=30, hight=90 and time_unit=second, the system will again execute the action
    randomly every 30 to 90 seconds after each execution.
    """
    time_unit:TimeUnit=TimeUnit.second
    low: int
    high: int

    def schedule_action(self, tag:str, action:Action, continuous:Continuous):
        callable = self.during_period(tag=tag, action=action)
        continuous.schedule_random_callable(
            tag=tag,
            callable=callable,
            time_unit=self.time_unit,
            low=self.low, high=self.high
        )



from whendo.core.action import Action
from whendo.core.resolver import resolve_action
from typing import List, Dict, Any, Optional
from enum import Enum
import json

class ExceptionAction(Action):
    exception_tag_field:str="excepted"

    def execute(self, tag:str=None, scheduler_info:dict=None):
        raise Exception(
            'purposely unsuccessful execution',
            self.json()
        )

class SuccessAction(Action):
    success_tag_field:str="successful"

    def execute(self, tag:str=None, scheduler_info:dict=None):
        return {
            'outcome':'purposely successful execution',
            'action':self.info()
            }

class NotAction(Action):
    operand:Action

    def execute(self, tag:str=None, scheduler_info:dict=None):
        operand_result = None
        try:
            operand_result = self.operand.execute(tag=tag, scheduler_info=scheduler_info)
        except Exception as exception:
            operand_result = exception
        if isinstance(operand_result, Exception):
            return {
                'outcome':'exception negated; action execution treated as a success',
                'exception':str(operand_result),
                'action':self.info()
                }
        else:
            raise Exception(
                'exception generated; action execution treated as a failure',
                self.json()
            )

class ListOpMode(str, Enum):
    ALL = 'all'
    OR = 'or'
    AND = 'and'

class ListAction(Action):
    op_mode:ListOpMode
    action_list:List[Action]
    exception_on_no_success:bool=False

    def execute(self, tag:str=None, scheduler_info:Dict[str,Any]=None):
        processing_count, success_count, failure_count, successful_actions, exception_actions = process_action_list(
            tag,
            scheduler_info,
            self.op_mode,
            action_list=self.action_list,
            successful_actions=[],
            exception_actions=[]
            )
        processing_info = {
            'processing_count':processing_count,
            'success_count':success_count,
            'exception_count':failure_count,
            'successful_actions':successful_actions,
            'exception_actions':exception_actions
        }
        if success_count == 0 and self.exception_on_no_success:
            raise Exception(
                'exception generated; action execution treated as a failure',
                json.dumps(self.dict()),
                json.dumps(processing_info)
            )
        else:
            return {'outcome':'list action executed', 'action':self.info(), 'processing_info':processing_info}

class IfElseAction(Action):
    op_mode:ListOpMode
    test_action:Action
    if_actions:List[Action]
    else_action:Action
    exception_on_no_success:bool=False

    def execute(self, tag:str=None, scheduler_info:dict=None):
        try:
            test_result = self.test_action.execute(tag=tag, scheduler_info=scheduler_info)
        except Exception as exception:
            test_result = exception
        processing_count = 1
        success_count = 0
        exception_count = 0
        successful_actions = []
        exception_actions = []
        else_test = isinstance(test_result, Exception)
        if else_test: # execute the else action
            exception_count = 1
            exception_actions.append(self.test_action.dict())

            try:
                else_result = self.else_action.execute(tag=tag, scheduler_info=scheduler_info)
            except Exception as exception:
                else_result = exception
            processing_count += 1
            if isinstance(else_result, Exception):
                exception_count += 1
                exception_actions.append(self.else_action.dict())
            else:
                success_count += 1
                successful_actions.append(self.else_action.dict())
        else:
            success_count += 1
            successful_actions.append(self.test_action.dict())
            processing_count, success_count, failure_count, successful_actions, exception_actions = process_action_list(
                tag=tag,
                scheduler_info=scheduler_info,
                op_mode=self.op_mode,
                action_list=self.if_actions,
                processing_count=processing_count,
                success_count=success_count,
                exception_count=exception_count,
                successful_actions=successful_actions,
                exception_actions=exception_actions
            )
        processing_info = {
            'else_test':else_test,
            'processing_count':processing_count,
            'success_count':success_count,
            'exception_count':exception_count,
            'successful_actions':successful_actions,
            'exception_actions':exception_actions
        }
        if success_count == 0 and self.exception_on_no_success:
            raise Exception(
                'exception generated; action execution treated as a failure',
                json.dumps(self.dict()),
                json.dumps(processing_info)
            )
        else:
            return {'outcome':'if else action executed', 'action':self.info(), 'processing_info':processing_info}

def process_action_list(
    tag:Optional[str],
    scheduler_info:Optional[Dict[str,Any]],
    op_mode:ListOpMode,
    action_list:List[Action],
    processing_count:int=0,
    success_count:int=0,
    exception_count:int=0,
    successful_actions:List[Dict[Any, Any]]=[],
    exception_actions:List[Dict[Any, Any]]=[]
    ):
    for action in action_list:
        try: # in case an Action does not return an exception
            result = action.execute(tag=tag, scheduler_info=scheduler_info)
        except Exception as exception:
            result = exception
        processing_count += 1
        if isinstance(result, Exception):
            exception_count += 1
            exception_actions.append(action.dict())
            if op_mode == ListOpMode.AND: # stop after first failure
                break
        else:
            success_count += 1
            successful_actions.append(action.dict())
            if op_mode == ListOpMode.OR: # stop after first success
                break
    return (processing_count, success_count, exception_count, successful_actions, exception_actions)
